/* 
 * Ice3 - DOR conduit
 * $Id: ds.c,v 1.1.1.1 2004/04/20 22:27:21 kaeld Exp $
 */

#include <arpa/inet.h>
#include <fcntl.h>
#include <getopt.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdio.h>
#include <string.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>


#define MAX_DOM_MSG_SIZE 400
#define D_IO_BUFSIZ 64000

char version[]	= "0.5";
short port	= 2121;
int  debug_level= 1;

char up[D_IO_BUFSIZ], down[D_IO_BUFSIZ];
int  uptr = 0,   dptr = 0;

char *optstring = "hp:v::";
struct option opts[] = {
	{ "help", no_argument, NULL, 'h' },
	{ "port", required_argument, NULL, 'p' },
	{ "debug", optional_argument, NULL, 'v' },
	{ NULL, 0, NULL, 0 }
};

void hexdump(FILE*, void*, int);

void show_usage() {
	fprintf(stderr, "usage :: ds [ -%s ] <dor-device>\n", optstring);
}

void show_help() {
	fprintf(stderr, "DOMServer %s\n", version);
	show_usage();
	fprintf(stderr,
		"options are -h or --help            this help\n"
		"            -p or --port=<port>     sets listening port\n"
		);
}

static int dread(int fd, void* buf, int* ptr)
{
	int nr, nrr;
	
	nr = D_IO_BUFSIZ - *ptr;
	if (nr > MAX_DOM_MSG_SIZE) nr = MAX_DOM_MSG_SIZE;
	nrr = read(fd, buf + *ptr, nr);
#ifndef NDEBUG
	if (debug_level > 4)
		fprintf(stderr, "Read %d bytes from FD (%d)\n", nrr, fd);
#endif
	if (nrr <= 0) return 1;
	*ptr += nrr;
	return 0;
}

static int dwrite(int fd, void* buf, int* ptr)
{
	int nw, nrw;

	nw = *ptr;
	if (nw > MAX_DOM_MSG_SIZE) nw = MAX_DOM_MSG_SIZE;
	nrw = write(fd, buf, nw);
#ifndef NDEBUG 
	if (debug_level > 4) 
		fprintf(stderr, "Wrote %d bytes to (%d)\n", nrw, fd);
	if (debug_level > 5)
		hexdump(stderr, buf, nrw);
#endif
	if (nrw <= 0) return 1;
	memmove(buf, buf + nrw, nrw);
	*ptr -= nrw;
	return 0;
}

int spinner(int s, int dh)
{
	int mx;
	int n;
	fd_set rfs, wfs;

	if (dh > s)
		mx = dh + 1;
	else
		mx = s + 1;
	
	FD_ZERO(&rfs); FD_ZERO(&wfs);

	/* Check buffers - add to select as needed */
	if (uptr < D_IO_BUFSIZ) FD_SET(dh, &rfs); 
	if (dptr) FD_SET(dh, &wfs); 
	if (dptr < D_IO_BUFSIZ) FD_SET(s, &rfs); 
	if (uptr) FD_SET(s, &wfs);
	
	n = select(mx, &rfs, &wfs, NULL, NULL);

	if (FD_ISSET(dh, &rfs)) dread(dh, up, &uptr);
	if (FD_ISSET(dh, &wfs)) dwrite(dh, down, &dptr);
	if (FD_ISSET(s, &rfs) && dread(s, down, &dptr)) return 1;
	if (FD_ISSET(s, &wfs) && dwrite(s, up, &uptr)) return 1;

	return 0;

}

int main(int argc, char **argv)
{
	int c;
	char* device;
	int dh, ss;
	struct sockaddr_in a, peer_addr;
	int padrlen; 

	while ((c = getopt_long(argc, argv, optstring, opts, NULL)) != EOF) {
		switch (c) {
		case 'h':
			show_help();
			exit(1);
		case 'p':
			port = atoi(optarg);
			break;
		case 'v':
			if (optarg)
				debug_level = atoi(optarg);
			else
				debug_level++;
			break;
		default:
			show_usage();
			exit(1);
		}
	}

	if (optind == argc) {
		show_usage();
		exit(1);
	}

	device = argv[optind++];

	dh = open(device, O_RDWR);
	if (dh < 0) {
		perror(argv[0]);
		exit(1);
	}
#ifndef NDEBUG
	if (debug_level > 2)
		fprintf(stderr, "opened device %s on FD %d\n", device, dh);
#endif

	ss = socket(PF_INET, SOCK_STREAM, 0);
	a.sin_family = AF_INET;
	a.sin_port = htons(port);
	a.sin_addr.s_addr = INADDR_ANY;
	if (bind(ss, (struct sockaddr*) &a, sizeof(a))) {
		perror(argv[0]);
		exit(1);
	}
	listen(ss, 0);
	padrlen = sizeof(peer_addr);
	while (1) {
		int sp;
		sp = accept(ss, (struct sockaddr*) &peer_addr, &padrlen);
#ifndef NDEBUG
		if (debug_level > 2) 
			fprintf(stderr, 
				"Accepted incoming connection from %s:%d\n",
				inet_ntoa(peer_addr.sin_addr),
				ntohs(peer_addr.sin_port));
#endif
		while (!spinner(sp, dh));
#ifndef NDEBUG
		if (debug_level > 2)
			fprintf(stderr, "Connection dropped\n");
#endif
		shutdown(sp, 2);
		close(sp);
	}

	shutdown(ss, 2);
	close(ss);
	close(dh);

	return 0;
}

	
